import asyncio
import inspect
import typing as t

from .types import AnyCallable


def get_function_name(func: AnyCallable) -> str:
    module_str = func.__module__
    if module_str == '__main__':
        module = inspect.getmodule(func)
        if module and module.__spec__:
            module_str = module.__spec__.name

    return f'{module_str}.{func.__name__}'


async def poll(step: float, timeout: float) -> t.AsyncIterator[float]:
    loop = asyncio.get_event_loop()
    start = loop.time()
    while True:
        before = loop.time()
        yield before - start
        after = loop.time()
        if after - start >= timeout:
            return
        wait = max([0, step - after + before])
        await asyncio.sleep(wait)



import typing as t
from collections import UserDict

import arq

from .types import AnyCallable

if t.TYPE_CHECKING:
    BaseRegistry = UserDict[str, arq.worker.Function]
else:
    BaseRegistry = UserDict


class Registry(BaseRegistry):

    def __init__(self) -> None:
        super().__init__()
        self.by_original_coro: t.Dict[AnyCallable, arq.worker.Function] = {}

    def __setitem__(
            self, arq_function_name: str, arq_function: arq.worker.Function,
    ) -> None:
        original_coroutine = t.cast(
            AnyCallable,
            arq_function.coroutine.__wrapped__,  # type: ignore
        )
        self.by_original_coro[original_coroutine] = arq_function
        self.data[arq_function_name] = arq_function

    def __delitem__(self, arq_function_name: str) -> None:
        arq_function = self.data[arq_function_name]
        original_coroutine = t.cast(
            AnyCallable,
            arq_function.coroutine.__wrapped__,  # type: ignore
        )
        del self.by_original_coro[original_coroutine]
        del self.data[arq_function_name]

    def add(self, arq_function: arq.worker.Function) -> None:
        self[arq_function.name] = arq_function

    def get_functions(self) -> t.Sequence[arq.worker.Function]:
        return tuple(self.values())

    def get_function_names(self) -> t.Sequence[str]:
        return tuple(arq_func.name for arq_func in self.values())



import asyncio
import logging
import typing as t

from arq.worker import async_check_health
from arq.worker import Function  # noqa: F401  need for reimport
from arq.worker import get_kwargs
from arq.worker import Worker as ArqWorker

from .app import Darq
from .types import JobCtx
from .utils import poll

logger = logging.getLogger('arq.worker')


class Worker(ArqWorker):

    def __init__(
            self, darq_app: Darq, queue: str,
            *args: t.Any, **kwargs: t.Any,
    ) -> None:
        self.darq_app = darq_app
        settings = self.get_settings(queue)

        self.warm_shutdown_timeout = 30
        self.warm_shutdown_task: t.Optional[asyncio.Task[None]] = None

        super().__init__(*args, **{**settings, **kwargs})

    def get_settings(self, queue: str) -> t.Dict[str, t.Any]:
        settings = {**self.darq_app.config}
        settings['functions'] = self.darq_app.registry.get_functions()
        if queue:
            settings['queue_name'] = queue

        on_startup = settings.get('on_startup')

        async def wrapped_on_startup(ctx: JobCtx) -> None:
            await self.darq_app.connect(ctx['redis'])
            if on_startup:
                await on_startup(ctx)

        on_shutdown = settings.get('on_shutdown')

        async def wrapped_on_shutdown(ctx: JobCtx) -> None:
            if on_shutdown:
                await on_shutdown(ctx)
            await self.darq_app.disconnect()

        settings['on_startup'] = wrapped_on_startup
        settings['on_shutdown'] = wrapped_on_shutdown
        return get_kwargs(settings)  # type: ignore

    def has_running_tasks(self) -> bool:
        return any(not task.done() for task in self.tasks)

    async def run_jobs(self, job_ids: t.Sequence[str]) -> None:
        if self.warm_shutdown_task:
            return
        await super().run_jobs(job_ids)  # type: ignore

    async def run_cron(self) -> None:
        if self.warm_shutdown_task:
            return
        await super().run_cron()  # type: ignore

    def handle_sig(self, signum: int) -> None:
        if self.warm_shutdown_task:
            self.warm_shutdown_task.cancel()
            super().handle_sig(signum)  # type: ignore
        else:
            self.warm_shutdown_task = self.loop.create_task(
                self.warm_shutdown(signum),
            )

    async def warm_shutdown(self, signum: int) -> None:
        if self.has_running_tasks():
            awaiting_task_count = sum(not task.done() for task in self.tasks)
            logger.info(
                'Warm shutdown. Awaiting for %d jobs with %d seconds timeout.',
                awaiting_task_count, self.warm_shutdown_timeout,
            )
            async for _ in poll(step=0.1, timeout=self.warm_shutdown_timeout):
                if not self.has_running_tasks():
                    break

        super().handle_sig(signum)  # type: ignore


def create_worker(darq: Darq, queue: str, **kwargs: t.Any) -> Worker:
    return Worker(darq, queue, **kwargs)


def run_worker(darq: Darq, queue: str, **kwargs: t.Any) -> Worker:
    worker = create_worker(darq, queue, **kwargs)
    worker.run()
    return worker


def check_health(darq: Darq) -> int:
    """
    Run a health check on the worker and return the appropriate exit code.
    :return: 0 if successful, 1 if not
    """
    loop = asyncio.get_event_loop()
    return loop.run_until_complete(
        async_check_health(
            darq.config.get('redis_settings'),
            darq.config.get('health_check_key'),
            darq.config.get('queue_name'),
        ),
    )



import logging.config
import os
import sys

import click
from arq.logs import default_log_config
from pydantic.utils import import_string

from .app import Darq
from .worker import check_health
from .worker import run_worker

health_check_help = 'Health Check: run a health check and exit.'
verbose_help = 'Enable verbose output.'


@click.command('darq')
@click.argument('darq-app', type=str, required=True)
@click.option('--check', is_flag=True, help=health_check_help)
@click.option('-v', '--verbose', is_flag=True, help=verbose_help)
@click.option('-Q', '--queue', type=str, default=None)
def cli(*, darq_app: str, check: bool, verbose: bool, queue: str) -> None:
    """
    Job queues in python with asyncio and redis.
    CLI to run the darq worker.

    DARQ_APP - path to Darq app instance.
    For example: someproject.darq.darq_app
    """
    sys.path.append(os.getcwd())
    darq = import_string(darq_app)
    if not isinstance(darq, Darq):
        raise click.BadArgumentUsage(
            f'DARQ_APP argument error. {darq!r} is not instance of {Darq!r}',
        )

    logging.config.dictConfig(default_log_config(verbose))

    if check:
        exit(check_health(darq))
    else:
        run_worker(darq, queue)


if __name__ == '__main__':
    cli()



import datetime
import sys
import typing as t

import arq

if sys.version_info >= (3, 8):
    from typing import TypedDict
else:
    from typing_extensions import TypedDict

AnyCallable = t.Callable[..., t.Any]
AnyTimedelta = t.Union[int, float, datetime.timedelta]


class JobCtx(TypedDict):
    redis: arq.ArqRedis
    job_id: str
    job_try: int
    enqueue_time: datetime.datetime
    score: int
    metadata: t.Dict[str, t.Any]



import datetime
import functools
import importlib
import typing as t

import arq
from arq.connections import ArqRedis
from arq.cron import CronJob
from arq.jobs import Job

from .registry import Registry
from .types import AnyCallable
from .types import AnyTimedelta
from .types import JobCtx
from .utils import get_function_name

TD_1_DAY = datetime.timedelta(days=1)


class DarqException(Exception):
    pass


class DarqConnectionError(DarqException):
    pass


class DarqConfigError(DarqException):
    pass


class Darq:

    def __init__(
            self,
            config: t.Dict[str, t.Any],
            default_job_expires: AnyTimedelta = TD_1_DAY,
            on_job_prerun: t.Optional[t.Callable[..., t.Any]] = None,
            on_job_postrun: t.Optional[t.Callable[..., t.Any]] = None,
            on_job_prepublish: t.Optional[t.Callable[..., t.Any]] = None,
    ) -> None:
        self.registry = Registry()
        self.default_job_expires = default_job_expires
        self.on_job_prerun = on_job_prerun
        self.on_job_postrun = on_job_postrun
        self.on_job_prepublish = on_job_prepublish
        self.config = config.copy()
        if 'functions' in self.config:
            raise DarqConfigError(
                '"functions" should not exist in config, all functions will '
                'be collected automatically. Just wrap your functions with '
                '@darq.task decorator.',
            )
        if 'queue_name' in self.config:
            raise DarqConfigError(
                '"queue_name" should not exist in config. '
                'To specify queue in worker - use "-Q" arg in cli.',
            )

        cron_jobs = self.config.pop('cron_jobs', [])
        self.config['cron_jobs'] = []
        self.add_cron_jobs(*cron_jobs)

        self.redis: t.Optional[arq.ArqRedis] = None
        if config.get('redis_pool'):
            self.redis = config['redis_pool']
        self.connected = bool(self.redis)

    async def connect(self, redis_pool: t.Optional[ArqRedis] = None) -> None:
        if self.connected:
            return

        if redis_pool:
            self.redis = redis_pool
        else:
            self.redis = await arq.create_pool(self.config['redis_settings'])
        self.connected = True

    async def disconnect(self) -> None:
        if not self.connected:
            return

        if self.redis:
            self.redis.close()
            await self.redis.wait_closed()
        self.connected = False

    def autodiscover_tasks(self, packages: t.Sequence[str]) -> None:
        for pkg in packages:
            importlib.import_module(pkg)

    def add_cron_jobs(self, *jobs: CronJob) -> None:
        for job in jobs:
            if not isinstance(job, CronJob):
                raise DarqException(f'{job!r} must be instance of CronJob')
            if job.coroutine not in self.registry.by_original_coro:
                raise DarqException(
                    f'{job.coroutine!r} is not registered. '
                    'Please, wrap it with @task decorator.',
                )
            # Replace original coroutine with wrapped by ``wrap_job_coroutine``
            arq_function = self.registry.by_original_coro[job.coroutine]
            job.coroutine = arq_function.coroutine  # type: ignore
            self.config['cron_jobs'].append(job)

    def wrap_job_coroutine(
            self, function: t.Callable[..., t.Any],
    ) -> t.Callable[..., t.Any]:

        @functools.wraps(function)
        async def wrapper(ctx: JobCtx, *args: t.Any, **kwargs: t.Any) -> t.Any:
            arq_function = self.registry.by_original_coro[function]
            ctx['metadata'] = kwargs.pop('__metadata__', {})
            if self.on_job_prerun:
                await self.on_job_prerun(ctx, arq_function, args, kwargs)

            result = await function(*args, **kwargs)

            if self.on_job_postrun:
                await self.on_job_postrun(
                    ctx, arq_function, args, kwargs, result,
                )
            return result

        return wrapper

    def task(
            self,
            func: t.Optional[AnyCallable] = None,
            *,
            keep_result: t.Optional[AnyTimedelta] = None,
            timeout: t.Optional[AnyTimedelta] = None,
            max_tries: t.Optional[int] = None,
            queue: t.Optional[str] = None,
            expires: t.Optional[AnyTimedelta] = None,
    ) -> t.Any:

        def _decorate(function: AnyCallable) -> AnyCallable:
            name = get_function_name(function)

            arq_function = arq.worker.func(
                coroutine=self.wrap_job_coroutine(function), name=name,
                keep_result=keep_result, timeout=timeout, max_tries=max_tries,
            )
            self.registry.add(arq_function)

            async def delay(*args: t.Any, **kwargs: t.Any) -> t.Optional[Job]:
                if queue and '_queue_name' not in kwargs:
                    kwargs['_queue_name'] = queue
                if '_expires' not in kwargs:
                    kwargs['_expires'] = expires or self.default_job_expires

                if not self.connected or not self.redis:
                    raise DarqConnectionError(
                        'Darq app is not connected. Please, make '
                        '"await <darq_instance>.connect()" before calling '
                        'this function',
                    )
                metadata: t.Dict[str, t.Any] = {}
                self.on_job_prepublish and await self.on_job_prepublish(
                    metadata, arq_function, args, kwargs,
                )
                if metadata:
                    kwargs['__metadata__'] = metadata
                return await self.redis.enqueue_job(name, *args, **kwargs)

            function.delay = delay  # type: ignore
            return function

        if func:
            return _decorate(func)

        return _decorate
